@@ -9,6 +9,7 @@
 #include <string>
 #include <sstream>
 #include <map>
+#include <set>
 #include <vector>
 #include <algorithm>
 #include <cctype>
@@ -53,8 +54,9 @@ class ldap_connection {
 
 public:
 	LDAP *connection;
+	bool bound;
 
-	ldap_connection() : connection(0) {}
+	ldap_connection() : connection(0), bound(false) {}
 	~ldap_connection() { disconnect(); }
 
 	bool connected() const { return connection != 0; }
@@ -80,6 +82,19 @@ public:
 
 	bool bind(const std::string &userid,
 		  const std::string &password)
+	{
+		if (do_bind(userid, password))
+		{
+			bound=true;
+			return true;
+		}
+
+		return false;
+	}
+
+private:
+	bool do_bind(const std::string &userid,
+		  const std::string &password)
 	{
 		std::vector<char> buffer(password.begin(), password.end());
 		struct berval cred;
@@ -87,10 +102,19 @@ public:
 		cred.bv_len=buffer.size();
 		cred.bv_val=&buffer[0];
 
-		return ok("ldap_sasl_bind_s",
-			  ldap_sasl_bind_s(connection, userid.c_str(),
-					   NULL, &cred,
-					   NULL, NULL, NULL));
+		if (connect() &&
+		    ok("ldap_sasl_bind_s",
+		       ldap_sasl_bind_s(connection, userid.c_str(),
+					NULL, &cred,
+					NULL, NULL, NULL)))
+			return true;
+
+		disconnect();
+		return connect() &&
+			ok("ldap_sasl_bind_s",
+			   ldap_sasl_bind_s(connection, userid.c_str(),
+					    NULL, &cred,
+					    NULL, NULL, NULL));
 	}
 };
 
@@ -418,6 +442,8 @@ bool ldap_connection::connect()
 {
 	if (connected()) return true;
 
+	bound=false;
+
 	DPRINTF("authldaplib: connecting to %s", authldaprc.ldap_uri.c_str());
 
 	if (ldapconncheck())
@@ -501,12 +527,13 @@ void ldap_connection::close()
 
 static int ldapopen()
 {
-	if (main_connection.connected()) return 0;
+	if (!main_connection.connected())
+	{
+		if (!main_connection.connect())
+			return 1;
+	}
 
-	if (!main_connection.connect())
-		return 1;
-
-	if (authldaprc.initbind)
+	if (authldaprc.initbind && !main_connection.bound)
 	{
 		/* Bind to server */
 		if (courier_authdebug_login_level >= 2)
@@ -539,7 +566,7 @@ class authldaprc_attributes {
 
 public:
 
-	std::map<std::string, std::string *> attributes;
+	std::map<std::string, std::vector<std::string *> > attributes;
 
 	std::string attribute(const char *name,
 			      const char *default_value,
@@ -550,7 +577,7 @@ public:
 		authldaprc.config(name, value, false, default_value);
 
 		if (!value.empty())
-			attributes[value]=&return_value;
+			attributes[value].push_back(&return_value);
 		return value;
 	}
 };
@@ -559,10 +586,11 @@ class authldaprc_attribute_vector : public std::vector<std::string> {
 
 public:
 
-	authldaprc_attribute_vector(const std::map<std::string, std::string *>
-				    &attributes)
+	authldaprc_attribute_vector(const std::map<std::string,
+				    std::vector<std::string *> > &attributes)
 	{
-		for (std::map<std::string, std::string *>::const_iterator
+		for (std::map<std::string, std::vector<std::string *> >
+			     ::const_iterator
 			     p=attributes.begin(); p != attributes.end(); ++p)
 		{
 			push_back(p->first);
@@ -581,6 +609,8 @@ public:
 	authldaprc_search_attributes(const std::vector<std::string> &attributes)
 		: copy_buffer(attributes)
 	{
+		std::set<std::string> dupes;
+
 		for (std::vector<std::string>::iterator
 			     p=copy_buffer.begin();
 		     p != copy_buffer.end(); ++p)
@@ -588,6 +618,9 @@ public:
 			if (p->empty())
 				continue;
 
+			if (dupes.find(*p) != dupes.end())
+				continue;
+			dupes.insert(*p);
 			p->push_back(0);
 			all_attributes_ptr.push_back(& (*p)[0]);
 		}
@@ -618,7 +651,8 @@ public:
 	{
 		struct timeval timeout_copy=timeout;
 
-		if (!conn.ok("ldap_search_ext_s",
+		if (!conn.connect() ||
+		    !conn.ok("ldap_search_ext_s",
 			     ldap_search_ext_s(conn.connection,
 					       basedn.c_str(),
 					       LDAP_SCOPE_SUBTREE,
@@ -630,6 +664,21 @@ public:
 					       100, &ptr)))
 		{
 			ptr=NULL;
+			conn.disconnect();
+			if (!conn.connect()
+			    || !conn.ok("ldap_search_ext_s",
+				     ldap_search_ext_s(conn.connection,
+						       basedn.c_str(),
+						       LDAP_SCOPE_SUBTREE,
+						       query.c_str(),
+						       search_attributes(),
+						       0,
+						       NULL, NULL,
+						       &timeout_copy,
+						       100, &ptr)))
+			{
+				ptr=NULL;
+			}
 		}
 	}
 
@@ -778,6 +827,21 @@ public:
 		return true;
 	}
 
+	bool operator()(const std::string &attrname,
+			const std::vector<std::string *> &values)
+	{
+		bool found=true;
+
+		for (std::vector<std::string *>::const_iterator
+			     b=values.begin();
+		     b != values.end(); ++b)
+		{
+			found=operator()(attrname, **b);
+		}
+
+		return found;
+	}
+
 	std::string options()
 	{
 		size_t i;
@@ -1161,10 +1225,10 @@ int authldap_lookup::operator()(int (*callback)(struct authinfo *, void *),
 
 	authldap_get_values get_value(main_connection.connection, entry, dn.c_str());
 
-	for (std::map<std::string, std::string *>::iterator
+	for (std::map<std::string, std::vector<std::string *> >::iterator
 		     p=attributes.begin(); p != attributes.end(); ++p)
 	{
-		get_value(p->first, *p->second);
+		get_value(p->first, p->second);
 	}
 
 	au=authldaprc.uid;
